// vim: sw=2 ts=2 expandtab

#include <SDL/SDL.h>
#include <SDL/SDL_image.h>
#include <SDL/SDL_rotozoom.h>

#include <pthread.h>
#include <semaphore.h>

#include <stdio.h>
#include <math.h>

#include "fft.h"
#include "utils.h"
#include "graphics.h"
#include "sound.h"

#include "config.h"
#include "enemies.h"
#include "powerups.h"

// Frames per second
#define FPS ((double)BUFFER_PARTS * SAMPLE_RATE / BLOCK_LEN)

// Number of new samples put into the buffer each frame
#define READ_SAMPLES (BLOCK_LEN / BUFFER_PARTS)

#define NUM_SHOTS 1000
#define NUM_ENEMIES 32

#define ENEMY_WEIGHTS_SUM 19
int enemyBar[19] = {
	0,
	1, 1, 1,
	2, 2, 2, 2, 2, 2, 2, 2,
	3, 3, 3, 3, 3,
	4, 4
};

struct Shot {
	int x, y;
	int height;
	int valid;
	double strength;
};

struct Player {
	int x, y;
  int points;
	double shotStrength;
	double shotPeriod;
	int shotHeight;
	SDL_Surface *sprite;
};

struct GameState {
  struct Shot shots[NUM_SHOTS];
  struct Enemy enemies[NUM_ENEMIES];
  struct Player player;
  struct Powerup powerup;

  double enemyScale;
};

double fft[BLOCK_LEN], smoothFFT[BLOCK_LEN], noiseEdge[BLOCK_LEN];
sem_t  fftSemaphore;

int running = 1;

const char toneNames[12][3] = {"C", "C#", "D", "D#", "E", "F", "F#", "G",
                               "G#", "A", "A#", "H"};

int barPos[5];
Mix_Chunk *barSample[5];
double curTime = 0;
double nextShotTime = 0;

char* freq_to_tone(uint32_t freq) {
	static char toneName[5];

	static const double baseFreq = 130.81278265029931733920; // C0
	int halftone = 120 + (int) (log(freq / baseFreq) / M_LN2 * 12. + 0.5);

	sprintf(toneName, "%-2s%2i", toneNames[halftone % 12], halftone / 12 - 10);

	return toneName;
}

void* fft_thread(void *param) {
	sample       buffer[BLOCK_LEN];
	sample       block[BLOCK_LEN];
	double       fftOutReal[BLOCK_LEN], fftOutImag[BLOCK_LEN];
	double       tmpFFT[BLOCK_LEN];
	double       fftAvg[2][BLOCK_LEN];
	double       fftSum, flickerNoiseIntensity;

	double       nextFrame;
	double       curTime;

	int i, j;

	init_fft();

	for(i = 0; i < BLOCK_LEN; i++) {
		tmpFFT[i] = 0;
		smoothFFT[i] = 10000;
	}

	while(running) {
		// shift the buffer left
		memmove(buffer, buffer + READ_SAMPLES,
		        (BLOCK_LEN - READ_SAMPLES) * sizeof(sample));

		size_t ret = fread(buffer + (BLOCK_LEN - READ_SAMPLES),
		                   sizeof(sample), READ_SAMPLES, stdin);
		if(ret != READ_SAMPLES) {
			break;
		}

		memcpy(block, buffer, BLOCK_LEN * sizeof(sample));

		apply_hanning(block);
		fft_transform(block, fftOutReal, fftOutImag);
		complex_to_absolute(fftOutReal, fftOutImag, tmpFFT);

		/* Calculate moving exponential averages in both directions */
		i = 0;
		j = BLOCK_LEN/2 - 1;
		fftAvg[0][i] = tmpFFT[i];
		fftAvg[1][j] = tmpFFT[j];
		for(i++, j--; i < BLOCK_LEN/2; i++, j--) {
			fftAvg[0][i] = fftAvg[0][i-1] * (1-FFT_NOISE_AVG_WEIGHT) + tmpFFT[i] * FFT_NOISE_AVG_WEIGHT;
			fftAvg[1][j] = fftAvg[0][j+1] * (1-FFT_NOISE_AVG_WEIGHT) + tmpFFT[j] * FFT_NOISE_AVG_WEIGHT;
		}

		sem_wait(&fftSemaphore);

		/* Sum over the frequencies in [100:10000] Hz, then use that to approximate the flicker noise intensity */
		fftSum = 0;
		for(i = 100L * BLOCK_LEN / SAMPLE_RATE; i < 10000 * BLOCK_LEN / SAMPLE_RATE; i++) {
			fftSum += tmpFFT[i];
		}
		flickerNoiseIntensity = fftSum / log(10000 / 100L);

		for(i = 0; i < BLOCK_LEN/2; i++) {
			/* Smooth the fft over time and frequency, and subtract flicker noise */
			smoothFFT[i] = FFT_NOISE_WEIGHT * i * fabs(fftAvg[0][i] + fftAvg[1][i] - tmpFFT[i] - flickerNoiseIntensity / i) + (1-FFT_NOISE_WEIGHT) * smoothFFT[i];

			fft[i] = fabs(tmpFFT[i] - flickerNoiseIntensity / i);
		}

#define LAPLACE_DISTANCE 5
		/* noiseEdge applies this kernel to smoothFFT: {1, 1, -4, 1, 1}, but uses a distance of LAPLACE_DISTANCE between the array elements.
		 * A minimum in noiseEdge indicates a peak in the FFT.
		 * */
		for(i = 0; i < LAPLACE_DISTANCE; i++) {
			noiseEdge[i] = smoothFFT[0] + smoothFFT[i/2] - 4 * smoothFFT[i] + smoothFFT[i*3/2] + smoothFFT[i*2];
			noiseEdge[BLOCK_LEN/2-i-1] = smoothFFT[BLOCK_LEN-1] - 2 * smoothFFT[BLOCK_LEN-i-1] + smoothFFT[BLOCK_LEN-i*2-1];
		}
		for(i = LAPLACE_DISTANCE; i < BLOCK_LEN/2 - LAPLACE_DISTANCE; i++) {
			noiseEdge[i] = smoothFFT[i-LAPLACE_DISTANCE] + smoothFFT[i-LAPLACE_DISTANCE/2] - 4 * smoothFFT[i] + smoothFFT[i+LAPLACE_DISTANCE/2] + smoothFFT[i+LAPLACE_DISTANCE];
		}

		double minEdge = HUGE_VAL;
		for(i = 0; i < BLOCK_LEN/2; i++) {
			if(noiseEdge[i] < minEdge)
				minEdge = noiseEdge[i];
		}
		minEdge--;
		for(i = 0; i < BLOCK_LEN; i++)
			noiseEdge[i] -= minEdge;

		sem_post(&fftSemaphore);

		curTime = get_hires_time();

		if(nextFrame < curTime - 0.1) {
			printf("Frame too early!\n");
			nextFrame = -1;
		}

		if(nextFrame < 0) {
			nextFrame = curTime + 0.05;
		}

		if(nextFrame > curTime + 0.1) {
			printf("Frame too late. Skipping.\n");
			nextFrame = curTime - 0.1;
		}

		nextFrame += 1.000/FPS;
		sleep_until(nextFrame);
	}

	return NULL;
}

SDL_Surface *loadSprite(char *path, char key_r, char key_g, char key_b) {
	SDL_Surface *tmp, *ret;
	tmp = IMG_Load(path);
	if(tmp == NULL) {
		fprintf(stderr, "Error loading sprite %s: %s\n", path, SDL_GetError());
		return NULL;
	}
	ret = SDL_DisplayFormatAlpha(tmp);
	SDL_FreeSurface(tmp);
	SDL_SetColorKey(ret, SDL_SRCCOLORKEY | SDL_RLEACCEL, SDL_MapRGB(ret->format, key_r, key_g, key_b));
	return ret;
}

int freqPos(double frequency, int *barPos)  {
	return barPos[0]
	  + (barPos[4] - barPos[0])
	    * (log(1. * frequency / MIN_DISPLAY_FREQ))
	    / (log(1. * MAX_DISPLAY_FREQ / MIN_DISPLAY_FREQ));
}

int freqSize(double frequency, int *barPos) {
	return 1 + freqPos(frequency, barPos) - freqPos(frequency + SAMPLE_RATE / BLOCK_LEN, barPos);
}

void applyPowerup(struct GameState *gameState, int type) {
	gameState->player.shotHeight = SHOT_HEIGHT * (NUM_POWERUP_TYPES - type);
	gameState->player.shotStrength = WHOLENOTE_STRENGTH * pow(SHOT_STRENGTH_FALLOFF, -type);
	gameState->player.shotPeriod = WHOLENOTE_FREQUENCY * pow(SHOT_FREQUENCY_FALLOFF, -type);
	nextShotTime = curTime;
}

void handleShots(SDL_Surface *screen, struct GameState *gameState, double timeDelta) {
	int i;
	for(i = 0; i < NUM_SHOTS; i++) {
		if(gameState->shots[i].valid) {
			gameState->shots[i].x += timeDelta * SCREEN_WIDTH / 5;
			if(gameState->shots[i].x > SCREEN_WIDTH)
				gameState->shots[i].valid = 0;

			SDL_Rect shotRect = {
				gameState->shots[i].x - SHOT_WIDTH/2,
				gameState->shots[i].y - gameState->shots[i].height/2,
				SHOT_WIDTH,
				gameState->shots[i].height};

			SDL_FillRect(screen, &shotRect, SDL_MapRGB(screen->format, 0x00, 0xFF, 0x00));
		}
	}
}
void handleEnemies(SDL_Surface *screen, struct GameState *gameState, double timeDelta) {
	int i;
	for(i = 0; i < NUM_ENEMIES; i++) {
		if(gameState->enemies[i].valid) {
			gameState->enemies[i].x -= timeDelta * SCREEN_WIDTH / 24 * (1 << gameState->enemies[i].type);
			if(gameState->enemies[i].x < 0) {
				// enemy got past gameState->player -> reduce score
				gameState->player.points -= gameState->enemies[i].hp;
				EnemyTypes[gameState->enemies[i].type].fails++;
				gameState->enemies[i].valid = 0;
				SDL_FreeSurface(gameState->enemies[i].sprite);
				gameState->enemies[i].sprite = NULL;
			}

			SDL_Rect enemyRect = {
				gameState->enemies[i].x - gameState->enemyScale * EnemyTypes[gameState->enemies[i].type].width/2,
				gameState->enemies[i].y - gameState->enemyScale * EnemyTypes[gameState->enemies[i].type].height/2,
				gameState->enemyScale * EnemyTypes[gameState->enemies[i].type].width,
				gameState->enemyScale * EnemyTypes[gameState->enemies[i].type].height};

      SDL_Surface *tmpSurf = zoomSurface(gameState->enemies[i].sprite, gameState->enemyScale, gameState->enemyScale, SMOOTHING_ON);
			SDL_BlitSurface(tmpSurf, NULL, screen, &enemyRect);
      SDL_FreeSurface(tmpSurf);
		}
	}
}
void handlePowerup(SDL_Surface *screen, struct GameState *gameState, double timeDelta) {
	if(gameState->powerup.type != PU_NONE) {
		gameState->powerup.x -= timeDelta * SCREEN_WIDTH / 8;
		SDL_Rect powerupRect = {
			gameState->powerup.x - PowerupTypes[gameState->powerup.type].width/2,
			gameState->powerup.y - PowerupTypes[gameState->powerup.type].height/2,
			PowerupTypes[gameState->powerup.type].width,
			PowerupTypes[gameState->powerup.type].height};
		SDL_BlitSurface(PowerupTypes[gameState->powerup.type].sprite, NULL, screen, &powerupRect);
	}
}
void handleNewEnemies(unsigned int *randomSeed, struct GameState *gameState){
	static double nextTime = 0;
	static int enemypos = 0;
	if(nextTime < curTime) {
		while(gameState->enemies[enemypos].valid)
			enemypos++;
		nextTime = curTime + ENEMY_MIN_DELAY + (rand_r(randomSeed) % ENEMY_DELAY_RANDVAL);

		int type = rand_r(randomSeed) % (NUM_ENEMY_TYPES/2)
			+ rand_r(randomSeed) % (NUM_ENEMY_TYPES/2 + (NUM_ENEMY_TYPES&1));

    int barIndex = enemyBar[rand_r(randomSeed)%ENEMY_WEIGHTS_SUM];
		gameState->enemies[enemypos].valid = 1;
		gameState->enemies[enemypos].sample = barSample[barIndex];
		gameState->enemies[enemypos].type = type;
		gameState->enemies[enemypos].x = SCREEN_WIDTH;
		gameState->enemies[enemypos].y = barPos[barIndex] + EnemyTypes[type].yoffset;
		gameState->enemies[enemypos].hp = EnemyTypes[type].hp;
		gameState->enemies[enemypos].sprite = SDL_ConvertSurface(
		 EnemyTypes[type].sprite,
		 EnemyTypes[type].sprite->format,
		 EnemyTypes[type].sprite->flags);

		enemypos++;
		if(enemypos >= NUM_ENEMIES)
			enemypos = 0;
	}
}
void handleNewPowerups(unsigned int *randomSeed, struct GameState *gameState){
	static double nextTime = 0;
	if(nextTime < curTime) {
		int r = rand_r(randomSeed)%9;
		int type = rand_r(randomSeed) % NUM_POWERUP_TYPES;
		nextTime = curTime + POWERUP_MIN_DELAY + (rand_r(randomSeed) % POWERUP_DELAY_RANDVAL);
		gameState->powerup.type = type;
		gameState->powerup.x = SCREEN_WIDTH;
		gameState->powerup.y = barPos[r/2] - (r&1) * (barPos[0]-barPos[1]) / 2;
	}
}
void handleNewShots(unsigned int *randomSeed, struct GameState *gameState){
	static int shotpos = 0;
	if(nextShotTime < curTime) {
		nextShotTime = curTime + gameState->player.shotPeriod;

		gameState->shots[shotpos].x = gameState->player.x;
		gameState->shots[shotpos].y = gameState->player.y;
		gameState->shots[shotpos].strength = gameState->player.shotStrength;
		gameState->shots[shotpos].height = gameState->player.shotHeight;
		gameState->shots[shotpos].valid = 1;
		shotpos++;
		if(shotpos >= NUM_SHOTS) {
			shotpos = 0;
		}
	}
}
void handleEnemyCollisions(struct GameState *gameState) {
	int i, j;
	for(i = 0; i < NUM_ENEMIES; i++) {
		for(j = 0; j < NUM_SHOTS; j++) {
			if(gameState->enemies[i].valid && gameState->shots[j].valid) {
				int distX = abs(gameState->enemies[i].x - gameState->shots[j].x);
				int distY = abs(gameState->enemies[i].y - gameState->shots[j].y);
				if(
				 distX < SHOT_WIDTH/2 + gameState->enemyScale * EnemyTypes[gameState->enemies[i].type].width/2 &&
				 distY < gameState->shots[j].height/2 + gameState->enemyScale * EnemyTypes[gameState->enemies[i].type].height/2
				 ) {
					gameState->shots[j].valid = 0;
					gameState->enemies[i].hp -= gameState->shots[j].strength;
					if(gameState->enemies[i].hp <= 0) {
						// enemy destroyed -> add original HP to score
						gameState->player.points += EnemyTypes[gameState->enemies[i].type].hp;
						SDL_FreeSurface(gameState->enemies[i].sprite);
						gameState->enemies[i].sprite = NULL;
						EnemyTypes[gameState->enemies[i].type].kills++;
						gameState->enemies[i].valid = 0;
						Mix_PlayChannel(-1, gameState->enemies[i].sample, 0);
					} else {
						double healthRatio = gameState->enemies[i].hp / EnemyTypes[gameState->enemies[i].type].hp;
						SDL_LockSurface(gameState->enemies[i].sprite);
						uint8_t *dst = gameState->enemies[i].sprite->pixels,
							*src = EnemyTypes[gameState->enemies[i].type].sprite->pixels;
						int pitch = gameState->enemies[i].sprite->pitch,
						    bpp = gameState->enemies[i].sprite->format->BytesPerPixel,
						    h = gameState->enemies[i].sprite->h,
						    w = gameState->enemies[i].sprite->w;
						for(int y = 0; y < h; y++) {
							for(int x = 0; x < w; x++) {
								int k = y * pitch + x * bpp;
								unsigned char col = src[k],
									      newcol = col * healthRatio;
								dst[k++] = newcol;
								dst[k++] = newcol;
								dst[k++] = col;
							}
						}
						SDL_UnlockSurface(gameState->enemies[i].sprite);
					}
				}
			}
		}
	}
}
void handlePowerupCollisions(struct GameState *gameState) {
	if(gameState->powerup.type != PU_NONE) {
		int distX = abs(gameState->powerup.x - gameState->player.x),
		    distY = abs(gameState->powerup.y - gameState->player.y);
		if(
		 distX < SHIP_WIDTH/2 + PowerupTypes[gameState->powerup.type].width/2 &&
		 distY < SHIP_HEIGHT/2 + PowerupTypes[gameState->powerup.type].height/2
		 ) {
			applyPowerup(gameState, gameState->powerup.type);
			gameState->powerup.type = PU_NONE;
		}
	}
}
int main(int argc, char **argv) {
	SDL_Surface *screen, *fftDisplay;
	SDL_Event    event;

	SDL_Rect fftDPSavRect,
		 fftDPSrcRect,
		 fftDPDstRect;

	uint32_t     white, black;

	double loudestFreqAvg = (MIN_DISPLAY_FREQ + MAX_DISPLAY_FREQ) / 2,
         avgColScale = 0;

  struct GameState gameState;
	memset(&gameState, 0, sizeof(struct GameState));

  gameState.enemyScale = 1.0;

	gameState.player = (struct Player){
    .x = 100,
    .y = 0,
    .points = 100,
    .shotStrength = 0,
    .shotPeriod = 0,
    .shotHeight = 0,
    .sprite = NULL};

	gameState.powerup = (struct Powerup){
		.type = PU_NONE,
		.x = 0,
		.y = 0 };

	double lastTime = curTime = get_hires_time();
	double timeDelta;
	double gameStartTime = curTime;

	double gameDuration;

	unsigned int randomSeed = (unsigned int) curTime;

	int i;
	int updateScreen = 1;

	int gameOver = 0;

	pthread_t fftThread;

	screen = init_graphics(SCREEN_WIDTH, SCREEN_HEIGHT);
	init_audio();

	// create semaphores
	sem_init(&fftSemaphore, 0, 1);

	// run the fft thread
	pthread_create(&fftThread, NULL, fft_thread, NULL);

	if(!screen) {
		return 1;
	}

	white = SDL_MapRGB(screen->format, 255, 255, 255);
	black = SDL_MapRGB(screen->format, 0, 0, 0);

	for(i = 0; i < NUM_POWERUP_TYPES; i++) {
		if(!(PowerupTypes[i].sprite = loadSprite(PowerupTypes[i].spriteFile, 0, 0, 0))) {
			exit(1);
		}
	}
	for(i = 0; i < NUM_ENEMY_TYPES; i++) {
		if(!(EnemyTypes[i].sprite = loadSprite(EnemyTypes[i].spriteFile, 0, 0, 0))) {
			exit(1);
		}
	}
	gameState.player.sprite = loadSprite("img/ship.png", 0, 0, 0);
	if(!gameState.player.sprite) {
		exit(1);
	}

	for(i = 0; i < 5; i++) {
		barPos[i] = SCREEN_HEIGHT - (i+1) * (SCREEN_HEIGHT - PLAYFIELD_START) / 6.;

    char filename[32];
    sprintf(filename, "snd/bar%i.ogg", i);
		barSample[i] = load_sample(filename);
  }

	fftDisplay = SDL_CreateRGBSurface(SDL_HWSURFACE, FFT_DISPLAY_SIZE-1, barPos[0]-barPos[4], 24, 0, 0, 0, 0);
	fftDPSavRect = (SDL_Rect){ 1, barPos[4], FFT_DISPLAY_SIZE-1, barPos[0]-barPos[4] },
	fftDPSrcRect = (SDL_Rect){ 0,         0, FFT_DISPLAY_SIZE-1, barPos[0]-barPos[4] },
	fftDPDstRect = (SDL_Rect){ 0, barPos[4], FFT_DISPLAY_SIZE-1, barPos[0]-barPos[4] };


	applyPowerup(&gameState, PU_THIRTYSECONDNOTE);

	while(running && !feof(stdin)) {
		curTime = get_hires_time();
		timeDelta = curTime - lastTime;
		lastTime = curTime;

		while(SDL_PollEvent(&event)) {
			if(event.type == SDL_QUIT || (event.type == SDL_KEYDOWN && event.key.keysym.sym == SDLK_ESCAPE)) {
				running = 0;
			}

			if(event.type == SDL_KEYDOWN && event.key.keysym.sym == SDLK_p) {
				updateScreen = !updateScreen;
			}

			if(event.type == SDL_KEYDOWN && event.key.keysym.sym == SDLK_r) {
        // restart the game
        gameOver = 0;
        gameState.player.points = 100;
        gameState.powerup.type = PU_NONE;
        gameState.enemyScale = 1;

        for(i = 0; i < NUM_ENEMIES; i++) {
          gameState.enemies[i].valid = 0;
        }

        for(i = 0; i < NUM_SHOTS; i++) {
          gameState.shots[i].valid = 0;
        }

        gameStartTime = get_hires_time();
			}
		}

		if(updateScreen) {
			sem_wait(&fftSemaphore);
			uint32_t loudestFreq = find_freq_with_max_snr(smoothFFT, noiseEdge, MIN_MOVEMENT_FREQ, MAX_MOVEMENT_FREQ);
			double maxSNR = get_snr_for_freq(fft, smoothFFT, loudestFreq);
			sem_post(&fftSemaphore);

			if(maxSNR > SNR_THRESHOLD) {
				if(fabs(loudestFreq - loudestFreqAvg) > SAMPLE_RATE / (double) BLOCK_LEN) {
					nextShotTime += 2. / FPS / 3;
				}
				loudestFreqAvg = loudestFreq * LOUDEST_FREQ_NEW_WEIGHT + loudestFreqAvg * (1 - LOUDEST_FREQ_NEW_WEIGHT);

				if(loudestFreqAvg > MAX_DISPLAY_FREQ) {
					loudestFreqAvg = MAX_DISPLAY_FREQ;
				} else if(loudestFreqAvg < MIN_DISPLAY_FREQ) {
					loudestFreqAvg = MIN_DISPLAY_FREQ;
				}
			}

			gameState.enemyScale *= pow(ENEMY_SHRINK_RATE, timeDelta);

			// clear screen, but save the old fftDisplay
			SDL_BlitSurface(screen, &fftDPSavRect, fftDisplay, NULL);
			SDL_Rect screenRect = {0, 0, SCREEN_WIDTH, SCREEN_HEIGHT};
			SDL_FillRect(screen, &screenRect, black);

			handleEnemies(screen, &gameState, timeDelta);
			handlePowerup(screen, &gameState, timeDelta);
			handleNewEnemies(&randomSeed, &gameState);
			handleNewPowerups(&randomSeed, &gameState);

      // do not handle certain events after the player died
			if(!gameOver) {
				handleNewShots(&randomSeed, &gameState);

        handleShots(screen, &gameState, timeDelta);

        // collision tests
        handleEnemyCollisions(&gameState);
        handlePowerupCollisions(&gameState);
			}

			if(gameState.player.points < 0 && !gameOver) {
				gameOver = 1;
				gameDuration = get_hires_time() - gameStartTime;
			}

			char str[64];

			// show the current loudest frequency
			SDL_Color textCol = {255, 255, 255};
			SDL_Color bgCol = {0, 0, 0};

			sprintf(str, "%-6.1f Hz (SNR: %5.1f dB)", loudestFreqAvg, 10*log10(maxSNR));
			draw_text(screen, "Frequency:", textCol, bgCol, 10, 10);
			draw_text(screen, str,          textCol, bgCol, 140, 10);
			draw_text(screen, "Tone:",      textCol, bgCol, 10, 40);
			draw_text(screen, freq_to_tone(loudestFreqAvg), textCol, bgCol, 140, 40);

			if(gameState.player.points > 0) {
				sprintf(str, "Points: %i", gameState.player.points);
				draw_text(screen, str, textCol, bgCol, SCREEN_WIDTH/2, 10);
			} else { // Game Over
				sprintf(str, "End of Game! You lasted %.2f seconds", gameDuration);
				draw_text(screen, str, textCol, bgCol, SCREEN_WIDTH/2, 10);
			}

			for(i = 0; i < 5; i++) {
				SDL_Rect barRect = { 64, barPos[i], SCREEN_WIDTH - 128, 2 };
				SDL_FillRect(screen, &barRect, white);
			}

			// show a dot representing the gameState.player’s position
			gameState.player.y =
				barPos[0]
				+ (barPos[4] - barPos[0])
				* (log(1. * loudestFreqAvg / MIN_DISPLAY_FREQ))
				/ (log(1. * MAX_DISPLAY_FREQ / MIN_DISPLAY_FREQ));

			if(gameState.player.y < PLAYFIELD_START) {
				gameState.player.y = PLAYFIELD_START;
			} else if(gameState.player.y > SCREEN_HEIGHT) {
				gameState.player.y = SCREEN_HEIGHT;
			}

			SDL_Rect endRect = { SCREEN_WIDTH - 64, 0, 64, SCREEN_HEIGHT };
			SDL_FillRect(screen, &endRect, black);

			if(gameState.player.points > 0) {
				SDL_Rect playerRect = {gameState.player.x - SHIP_WIDTH/2, gameState.player.y - SHIP_HEIGHT/2, SHIP_WIDTH, SHIP_HEIGHT};
				SDL_BlitSurface(gameState.player.sprite, NULL, screen, &playerRect);
			}

			double maxnoise = 0,
			       maxsignal = fft[(int)((loudestFreqAvg>=SAMPLE_RATE?SAMPLE_RATE/2:loudestFreqAvg)*BLOCK_LEN/SAMPLE_RATE)];

			for(i = MIN_DISPLAY_FREQ * BLOCK_LEN / SAMPLE_RATE; i < MAX_DISPLAY_FREQ * BLOCK_LEN / SAMPLE_RATE; i++) {
				if(smoothFFT[i] > maxnoise)
					maxnoise = smoothFFT[i];
			}


      avgColScale = 0.99 * avgColScale + 0.01 * 255. / maxsignal * noiseEdge[(int)(BLOCK_LEN*loudestFreqAvg/SAMPLE_RATE)];
			double colScale[3] = {
				avgColScale > 255. / maxsignal * noiseEdge[(int)(BLOCK_LEN*loudestFreqAvg/SAMPLE_RATE)] ? 255. / maxsignal * noiseEdge[(int)(BLOCK_LEN*loudestFreqAvg/SAMPLE_RATE)] : avgColScale,
				255. / maxnoise,
				255. };
			for(i = MIN_DISPLAY_FREQ * BLOCK_LEN / SAMPLE_RATE; i < MAX_DISPLAY_FREQ * BLOCK_LEN / SAMPLE_RATE; i++) {
				int col = colScale[0] * fft[i] / noiseEdge[i];
        SDL_Rect barRect = {
          FFT_DISPLAY_SIZE-1, freqPos(i * SAMPLE_RATE / BLOCK_LEN, barPos),
					 1, freqSize(i * SAMPLE_RATE / BLOCK_LEN, barPos)
				};
				SDL_FillRect(screen, &barRect, SDL_MapRGB(fftDisplay->format, col, col, col));
				col = colScale[1] * smoothFFT[i];
				barRect = (SDL_Rect){
					FFT_DISPLAY_SIZE, freqPos(i * SAMPLE_RATE / BLOCK_LEN, barPos),
					16, freqSize(i * SAMPLE_RATE / BLOCK_LEN, barPos)
				};
				SDL_FillRect(screen, &barRect, SDL_MapRGB(screen->format, col, col, col));
			}
			SDL_BlitSurface(fftDisplay, &fftDPSrcRect, screen, &fftDPDstRect);
			SDL_Rect barRect = (SDL_Rect){
				FFT_DISPLAY_SIZE+16, freqPos(loudestFreq, barPos),
				8, freqSize(loudestFreq, barPos)
			};
			SDL_FillRect(screen, &barRect, SDL_MapRGB(screen->format, 0, 255, 0));
			barRect = (SDL_Rect){
				FFT_DISPLAY_SIZE+24, freqPos(loudestFreqAvg, barPos),
				8, freqSize(loudestFreqAvg, barPos)
			};
			SDL_FillRect(screen, &barRect, SDL_MapRGB(screen->format, 0, 255, 255));

			SDL_UpdateRect(screen, 0, 0, 0, 0);
			//SDL_Flip(screen);
		}
	}

	pthread_join(fftThread, NULL);

	SDL_Quit();

	return 0;
}
